.common-vars {
  content: "common-vars"; }

.common-struct {
  content: "common-struct"; }

.myvars {
  content: "1"; }

.a1 {
  content: "a1"; }

.a2 {
  content: "a2"; }

.common-vars {
  content: "common-vars"; }

.common-struct {
  content: "common-struct"; }

.myvars {
  content: "2"; }

.b1 {
  content: "b1"; }

.b2 {
  content: "b2"; }

#the-last {
  content: "LAST"; }